//# componentlist_cmpt.cc...interface class for componentlist tools 
//# Copyright (C) 2007
//# Associated Universities, Inc. Washington DC, USA.
//#
//# This program is free software; you can redistribute it and/or modify it
//# under the terms of the GNU General Public License as published by the Free
//# Software Foundation; either version 2 of the License, or (at your option)
//# any later version.
//#
//# This program is distributed in the hope that it will be useful, but WITHOUT
//# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
//# more details.
//#
//# You should have received a copy of the GNU General Public License along
//# with this program; if not, write to the Free Software Foundation, Inc.,
//# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
//#
//# Correspondence concerning AIPS++ should be addressed as follows:
//#        Internet email: casa-feedback@nrao.edu.
//#        Postal address: AIPS++ Project Office
//#                        National Radio Astronomy Observatory
//#                        520 Edgemont Road
//#                        Charlottesville, VA 22903-2475 USA
//#
//# $Id$

/***
 * Framework independent implementation file for componentlist...
 *
 * Implement the componentlist component here.
 * 
 * // TODO: WRITE YOUR DESCRIPTION HERE! 
 * Need to implement the error in position and flux for each component
 * 
 * 
 ***/

#include <iostream>
#include <componentlist_cmpt.h>
#include <components/ComponentModels/ComponentList.h>
#include <components/ComponentModels/ComponentShape.h>
#include <components/ComponentModels/ConstantSpectrum.h>
#include <components/ComponentModels/SpectralModel.h>
#include <components/ComponentModels/PowerLogPoly.h>
#include <components/ComponentModels/SpectralIndex.h>
#include <components/ComponentModels/TabularSpectrum.h>
#include <casacore/casa/Exceptions/Error.h>
#include <casacore/casa/BasicSL/STLIO.h>
#include <casacore/casa/Logging/LogOrigin.h>
#include <casacore/casa/Logging/LogIO.h>
#include <casacore/casa/Containers/Record.h>
#include <casacore/casa/Arrays/IPosition.h>
#include <casacore/casa/Arrays/Vector.h>
#include <casacore/casa/Arrays/ArrayMath.h>
#include <casacore/casa/BasicSL/String.h>
#include <casacore/casa/Utilities/MUString.h>
#include <casacore/casa/Quanta/Quantum.h>
#include <casacore/casa/Quanta/QuantumHolder.h>
#include <casacore/casa/Quanta/Unit.h>
#include <casacore/casa/Quanta/MVAngle.h>
#include <casacore/casa/Quanta/MVDirection.h>
#include <casacore/casa/namespace.h>
#include <casacore/casa/OS/Path.h>
#include <components/ComponentModels/ComponentList.h>
#include <components/ComponentModels/Flux.h>
#include <components/ComponentModels/SkyComponent.h>
#include <casacore/measures/Measures/MeasureHolder.h>

using namespace std;

using namespace casacore;
using namespace casa;

namespace casac {

componentlist::componentlist() 
{
      itsLog = new LogIO();
      itsList = new ComponentList();
      itsBin = new ComponentList();
}

componentlist::~componentlist()
{
   if(itsLog){
      delete itsLog;
      itsLog = 0;
   }
   if(itsList){
      delete itsList;
      itsList = 0;
   }
   if(itsBin){
      delete itsBin;
      itsBin = 0;
   }
}

bool componentlist::open(const std::string& filename, const bool nomodify,
                         const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "open"));

  bool rstat(false);
  try {
    if(!(Table::isReadable(filename)) || !(Table(filename).tableDesc().isColumn("Shape_Parameters"))){
      throw(AipsError(String(filename)+" is non existant or is not a componentlist table"));
    }
    if(itsList !=0) 
      delete itsList;
    if(itsBin !=0) 
      delete itsBin;
    if(filename.size() > 0){
      itsList = new ComponentList(Path(filename), nomodify);
      itsBin = new ComponentList();
    } else {
      itsList = new ComponentList();
      itsBin = new ComponentList();
    }
    rstat=true;
  } catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

long componentlist::asciitocomponentlist(const std::string& /*filename*/,
                                        const std::string& /*asciifile*/,
                                        const std::string& /*refer*/,
                                        const std::string& /*format*/,
                                        const ::casac::record& /*direction_*/,
                                        const ::casac::record& /*spectrum*/,
                                        const ::casac::record& /*flux*/, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "asciitocomponentlist"));

  // TODO : IMPLEMENT ME HERE !
  long rstat(0);
  try{
    if(itsList && itsBin)
      *itsLog << LogIO::WARN
              << "asciitocomponentlist not implemented yet" << LogIO::POST;
    else
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::concatenate(const ::casac::variant& list,
                                const std::vector<long>& which, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "concatenate"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      Vector<Int> what(which);
      ComponentList toBeAdded;
      if(list.type() == ::casac::variant::STRING ||
         list.type() == ::casac::variant::STRINGVEC){
        String filename(list.toString());
        if(!Table::isReadable(filename)){
          throw(AipsError("Cannot read componentlist "+filename));
        }
        toBeAdded=ComponentList(Path(filename), true);
      }
      else if(list.type()==::casac::variant::RECORD){
        ::casac::variant localvar(list);
        Record * ptrRec = toRecord(localvar.asRecord());
        String error;
        if(!(toBeAdded.fromRecord(error, *ptrRec))){
          delete ptrRec;
          throw(AipsError(String("Error ") + error +
                          String(" in converting from record")));
        }
        delete ptrRec;
      }
      else{
        *itsLog << LogIO::SEVERE
                << "Can concatenate only componentlists which are on disk or are a record for now "
                << LogIO::POST;
        return false;
      }
      if(toBeAdded.nelements() < 1){
        *itsLog << LogIO::WARN << "Empty componentlist" << LogIO::POST;
        return false;
      }
      if(what.nelements()==0 || what[0] < 0){
        what.resize(toBeAdded.nelements());
        indgen(what);
      }
      for (uInt k=0; k < what.nelements(); ++k){
        if(uInt(what[k]) < toBeAdded.nelements()){
          itsList->add(toBeAdded.component(k));
        }
        else{
          *itsLog << LogIO::SEVERE 
                  << "component " << k << "does not exist in this list "
                  << LogIO::POST;
        }
      } 
      return true;
    }
    else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool
componentlist::remove(const std::vector<long>& which, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "remove"));

  bool rstat(false);
  try{
    std::vector<int> intwhich(which.begin( ),which.end( ));
    if(itsList && itsBin){
      const Vector<int> intVec(_checkIndices(intwhich, "remove",
                                             "No components removed"));
      for(uInt c = 0; c < intVec.nelements(); c++)
        itsBin->add(itsList->component(intVec(c)));
      itsList->remove(intVec);
      rstat = true;
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::purge()
{
  itsLog->origin(LogOrigin("componentlist", "purge"));

  bool rstat(false);
  try {
        if(itsList && itsBin){
           Vector<Int> indices(itsBin->nelements());
           indgen(indices);
           itsBin->remove(indices);
	} else {
	  *itsLog << LogIO::WARN << "componentlist is not opened, please open first" << LogIO::POST;
	}
  } catch (AipsError& x){
	  *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
  }
  return rstat;
}

bool componentlist::recover(const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "recover"));

  bool rstat(false);
  try {
        if(itsList && itsBin){
       uInt i = itsBin->nelements();
       while (i > 0) {
              i--;
              itsList->add(itsBin->component(i));
       }
       purge();
	} else {
	  *itsLog << LogIO::WARN << "componentlist is not opened, please open first" << LogIO::POST;
	}

  } catch (const AipsError& x){
	  *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
  }
  return rstat;
}

long componentlist::length()
{
  itsLog->origin(LogOrigin("componentlist", "length"));

  int rstat(0);
  try{
    if(itsList && itsBin){
      rstat = itsList->nelements();	  
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::vector<long> componentlist::indices()
{
  itsLog->origin(LogOrigin("componentlist", "indices"));

  // TODO : IMPLEMENT ME HERE !
  std::vector<long> rstat(0);
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "indices not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch(AipsError x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::sort(const std::string& criteria, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "sort"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      ComponentList::SortCriteria sortEnum = ComponentList::type(criteria);
      if (sortEnum == ComponentList::UNSORTED) {
        *itsLog << "Bad sort criteria." << endl
               << "Allowed values are: 'flux', 'position' & 'polarization'"
               << endl << "No sorting done."
               << LogIO::EXCEPTION;
      }
      itsList->sort(sortEnum);
      return true;
    } else {
      *itsLog << LogIO::WARN << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
	  *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
  }
  return rstat;
}

bool componentlist::isphysical(const std::vector<long>& which)
{
  itsLog->origin(LogOrigin("componentlist", "isphysical"));

  bool rstat(false);
  try{
    std::vector<int> intwhich(which.begin( ),which.end( ));
    if(itsList && itsBin){
      const Vector<Int> intVec(_checkIndices(intwhich, "is_physical",
                                             "Not checking any components"));
      rstat = itsList->isPhysical(intVec);
    }
    else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::vector<double> componentlist::sample(const ::casac::variant& /*direction_*/, 
                                          const ::casac::variant& /*pixellatsize*/, 
                                          const ::casac::variant& /*pixellongsize*/, 
                                          const ::casac::variant& /*frequency*/)
{
  itsLog->origin(LogOrigin("componentlist", "sample"));

  // TODO : IMPLEMENT ME HERE !
  std::vector<double> rstat(0);
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "sample not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::rename(const std::string& filename, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "rename"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      itsList->rename(Path(filename), Table::NewNoReplace);
      rstat=true;
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::simulate(const long howmany, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "simulate"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      for (int k=0; k < howmany; ++k){
        itsList->add(SkyComponent());
      }	  
      rstat=true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::addcomponent(
    const ::casac::variant& flux, const std::string& fluxunit,
    const std::string& polarization, const ::casac::variant& dir, 
    const std::string& shape, const ::casac::variant& majoraxis,
    const ::casac::variant& minoraxis, const ::casac::variant& positionangle,
    const ::casac::variant& freq, const std::string& spectrumtype,
    const ::casac::variant& index, const std::vector<double>& /*optionalparms*/,
    const std::string& label
) {
    itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
    try{
        if (! (itsList && itsBin)) {
            *itsLog << LogIO::WARN
                << "componentlist is not opened, please open first" << LogIO::POST;
            return false;
        }
        simulate(1);
        int which = itsList->size() - 1;
        setlabel(which, label, true);
        ::casac::variant error;
        setflux(which, flux, fluxunit, polarization, error, true);
        casacore::MDirection theDir;
        unique_ptr<::casac::variant> tmpdir(
            dir.toString() == "[]"
            ? new ::casac::variant(std::string("J2000 00h00m00.0 90d00m00"))
            : new ::casac::variant(dir)
        );
        ThrowIf(
            ! casaMDirection(*tmpdir, theDir),
            "Could not interpret direction parameter" 
        );
        MVDirection newDir = theDir.getValue();
        const Vector<Int> intVec(
            _checkIndices(
                which, __FUNCTION__,
                "Direction not changed on any components"
            )
        );
        itsList->setRefDirection(intVec, newDir);
        setrefdirframe(which, theDir.getRefString(), true);
        ::casac::variant majoraxiserror; 
        ::casac::variant minoraxiserror; 
        ::casac::variant positionangleerror;
        setshape(
            which, shape, majoraxis, minoraxis, positionangle,
            majoraxiserror,minoraxiserror, positionangleerror
        );
        setspectrum(which, spectrumtype, index);
        MFrequency theFreq;
        unique_ptr<::casac::variant> tmpfreq(
            freq.toString() == "[]"
            ? new ::casac::variant(std::string("LSRK 1.420GHz"))
            : new ::casac::variant(freq)
        );
        ThrowIf(
            ! casaMFrequency(*tmpfreq, theFreq),
            "Could not interpret frequency parameter" 
        );
        setfreq(which, theFreq.get("GHz").getValue(), "GHz", true);
        setfreqframe(which, theFreq.getRefString(), true);
        return true;
    }
    catch (const AipsError& x){
        //exception is thrown so the last component is bad ...lets remove it
        Vector<Int> remov(1,0);
        if (itsList->nelements() > 0){
            remov[0] = itsList->size() - 1;
            itsList->remove(remov);
        }
        *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
        RETHROW(x)
    }
    return false;
}

bool componentlist::close(const bool log)
{
  bool rstat(false);
  try {
      rstat = done();
	  if(log) {
          *itsLog << LogOrigin(LogOrigin("componentlist", __FUNCTION__))
          << LogIO::WARN << "componentlist closed" << LogIO::POST;
      }
	  rstat=true;
  } catch (const AipsError& x){
	  *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
  }
  return rstat;
}

  bool componentlist::edit(const long /*which*/, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "edit"));

  // TODO : IMPLEMENT ME HERE !
  bool rstat(false);
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "edit  not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (AipsError x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::done()
{
  itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
  bool rstat(false);
  try{
    if(itsList) {
        delete itsList;
        itsList = nullptr;
    }
       
    if(itsBin) {
        delete itsBin;
        itsBin = nullptr;
    }
    //bring it back to the state of construction
    itsList = new ComponentList();
    itsBin = new ComponentList();
    rstat=true;
  }
  catch(const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::haskeyword(const string& keyword) {
  itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
    try {
        return itsList->hasKeyword(String(keyword));
    }
    catch (const AipsError& x){
	    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	    RETHROW(x)
    }
    return false;
}

void componentlist::putkeyword(const string& keyword, const variant& value) {
  itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
    try {
        itsList->putKeyword(String(keyword), value);
    }
    catch (const AipsError& x){
	    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	    RETHROW(x)
    }
}

::casac::variant* componentlist::getkeyword(const std::string& keyword) {
  itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
    try {
        return itsList->getKeyword(String(keyword));
    }
    catch (const AipsError& x){
	    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	    RETHROW(x)
    }
    return nullptr;
}

bool componentlist::select(const std::vector<long>& which)
{
  itsLog->origin(LogOrigin("componentlist", "select"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      std::vector<int> intwhich(which.begin( ),which.end( ));
      const Vector<Int> intVec(_checkIndices(intwhich, "select",
                                             "No components selected"));
      itsList->select(intVec);
      rstat=true;
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::deselect(const std::vector<long>& which)
{
  itsLog->origin(LogOrigin("componentlist", "deselect"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      std::vector<int> intwhich(which.begin( ),which.end( ));
      const Vector<Int> intVec(_checkIndices(intwhich, "deselect",
                                             "No components deselected"));
      itsList->deselect(intVec);
       rstat=true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::vector<long> componentlist::selected()
{
  itsLog->origin(LogOrigin("componentlist", "selected"));

  try{
    if(itsList && itsBin){
      Vector<Int> theChosen = itsList->selected();
      return std::vector<long>(theChosen.begin( ),theChosen.end( ));
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return std::vector<long>(0);
}

std::string componentlist::getlabel(const long which)
{
  itsLog->origin(LogOrigin("componentlist", "getlabel"));

  std::string rstat("");
  try{
    if(itsList && itsBin){
      _checkIndex(which);
      rstat = itsList->component(which).label().c_str();
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::setlabel(const long which, const std::string& value,
                             const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "setlabel"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      itsList->setLabel(Vector<Int>(1, which) , value);
      rstat=true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::vector<double> componentlist::getfluxvalue(const long which)
{
  itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
  std::vector<double> rstat(0);
  try{
    if(itsList && itsBin){
      _checkIndex(which);
      Vector<DComplex> fluxes = itsList->component(which).flux().value();

      Vector<Double> realfluxes(fluxes.nelements());
      for (uInt k=0; k < fluxes.nelements(); ++k){
        realfluxes[k]=real(fluxes[k]);
      }
      realfluxes.tovector(rstat);
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::string componentlist::getfluxunit(const long which)
{
  itsLog->origin(LogOrigin("componentlist", "getfluxunit"));

  std::string rstat("");
  try{
    if(itsList && itsBin){
      _checkIndex(which);
      rstat=itsList->component(which).flux().unit().getName();
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::string componentlist::getfluxpol(const long /*which*/)
{
  itsLog->origin(LogOrigin("componentlist", "getfluxpol"));

  // TODO : IMPLEMENT ME HERE !
  std::string rstat("");
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "getfluxpol not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::vector<double> componentlist::getfluxerror(const long which) {
	itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
	try{
		ThrowIf(
			! itsList || ! itsBin,
			"componentlist is not opened, please open first"
		);
		_checkIndex(which);
		Vector<DComplex> fluxes = itsList->component(which).flux().errors();
		vector<Double> realfluxes;
		for(DComplex flux: fluxes) {
	        realfluxes.push_back(real(flux));
		}
		return realfluxes;
	}
	catch (const AipsError& x){
		*itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	    RETHROW(x)
	}
	return vector<double>();
}

bool componentlist::setflux(
	const long which, const ::casac::variant& varvalue,
	const std::string& unit,
	const std::string& polarization,
	const ::casac::variant& error, const bool /*log*/
) {
	itsLog->origin(LogOrigin("componentlist", "setflux"));

	bool rstat(false);
	try{
		ThrowIf(
			! itsList || ! itsBin,
			"componentlist is not opened, please open first"
		);
		Flux<Double> newFlux;
		const ComponentType::Polarisation pol = (ComponentType::Polarisation)(_checkFluxPol(polarization));
		newFlux.setPol(pol);

		const Unit fluxUnit(unit);
		if (fluxUnit != Unit("Jy")) {
			*itsLog << "The flux units must have the same dimensions as the Jansky"
				<< endl << "Flux not changed on any components"
				<< LogIO::EXCEPTION;
		}
		newFlux.setUnit(fluxUnit);
		//Deal with the value...complex for nothing for now
		Vector<Complex> value;
		if( String(varvalue.toString()) == String("[]") ||
		    (varvalue.type()== ::casac::variant::INTVEC) ||
		    (varvalue.type()==::casac::variant::INT) ||
		    (varvalue.type()== ::casac::variant::DOUBLEVEC) ||
		    (varvalue.type()==::casac::variant::DOUBLE) ||
		    (varvalue.type()== ::casac::variant::COMPLEXVEC) ||
		    (varvalue.type()==::casac::variant::COMPLEX)
		) {  
		  Vector<DComplex> tmpintvec= varvalue.toComplexVec().size() > 0 ? Vector<DComplex>(varvalue.toComplexVec()): Vector<DComplex>(4, 0.0);
			value.resize(tmpintvec.size());
			convertArray(value, tmpintvec);
		}
		else{
			throw(AipsError("Could not understand the type of flux value variable"));
		}
		variant::TYPE errorType = error.type();
		Vector<DComplex> errs;
		if(
			(errorType == ::casac::variant::INTVEC) ||
			(errorType == ::casac::variant::INT) ||
			(errorType == ::casac::variant::DOUBLEVEC) ||
			(errorType ==::casac::variant::DOUBLE) ||
			(errorType == ::casac::variant::COMPLEXVEC) ||
			(errorType ==::casac::variant::COMPLEX)
		) {
			Vector<DComplex> tmpError(error.toComplexVec());
			ThrowIf(
				tmpError.size() != value.size(),
				"Error array must have same length as value array"
			);
			errs.resize(tmpError.size());
			convertArray(errs, tmpError);
		}

		if ((value.size() == 1) &&  (pol == ComponentType::STOKES)) {
			newFlux.setValue(real(value[0]));
			if (errs.size() == 1) {
				newFlux.setErrors(errs);
			}
		}
		else if (value.size() == 4) {
			Vector<Double> tmpC(4);
			Vector<Double> tmpE(errs.size());
			for(int i=0;i<4;i++) {
				tmpC[i] = real(value[i]);
				if (tmpE.size() == 4) {
					tmpE[i] = real(errs[i]);
				}
			}
			newFlux.setValue(tmpC);
			if (tmpE.size() == 4) {
				newFlux.setErrors(tmpE[0], tmpE[1], tmpE[2], tmpE[3]);
			}
		}
		else {
			*itsLog << "The flux must have one or four elements," << endl
				<< "one element can only be used if the polarization is 'Stokes'."
				<< endl << "Flux not changed on any components"
				<< LogIO::EXCEPTION;
		}
		const Vector<Int> intVec(_checkIndices(which, "setflux",
				"Flux not changed on any components"));
		itsList->setFlux(intVec, newFlux);
		rstat = true;

	}
	catch (const AipsError& x){
		*itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
		RETHROW(x)
	}
	return rstat;
}

bool componentlist::convertfluxunit(const long which, const std::string& unit)
{
  itsLog->origin(LogOrigin("componentlist", "convertfluxunit"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      const Unit fluxUnit(unit);
      if (fluxUnit != Unit("Jy")) {
        *itsLog << LogIO::SEVERE
                << "The flux units must have the same dimensions as the Jansky"
                << endl << "Flux not changed on any components"
                << LogIO::POST;
        return false;
      }
      const Vector<Int> intVec(_checkIndices(which, "convertfluxunit",
                                             "Flux not changed on any components"));
      itsList->convertFluxUnit(intVec, fluxUnit);
      rstat = true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::convertfluxpol(const long which, const std::string& polarization)
{
  itsLog->origin(LogOrigin("componentlist", "convertfluxpol"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      const Vector<Int> intVec(_checkIndices(which, "convertfluxunit",
                                             "Flux not changed on any components"));
      itsList->convertFluxPol(intVec,
                       (ComponentType::Polarisation)_checkFluxPol(polarization));
      rstat=true;
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch(const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

::casac::record* componentlist::getrefdir(const long which)
{
  itsLog->origin(LogOrigin("componentlist", "getrefdir"));

  ::casac::record *retval = 0;
  try{
    if(itsList && itsBin){
      _checkIndex(which);
      casacore::MDirection refdir = itsList->component(which).shape().refDirection();
      ostringstream oss;
      refdir.print(oss);
      *itsLog << LogIO::NORMAL3 << String(oss) << LogIO::POST;
      MeasureHolder dirhold(refdir);
      Record outRec;
      String error;
      if (dirhold.toRecord(error, outRec)) {
        retval = fromRecord(outRec);
      } else {
        error += String("....Failed coversion of direction to record.\n");
        throw(AipsError(error));
      }
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return retval;
}

std::string componentlist::getrefdirra(const long /*which*/, const std::string& /*unit*/,
                                       const long /*precision*/)
{
  itsLog->origin(LogOrigin("componentlist", "getrefdirra"));

  // TODO : IMPLEMENT ME HERE !
  std::string rstat("");
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "getrefdirra not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::string componentlist::getrefdirdec(const long /*which*/, const std::string& /*unit*/,
                                        const long /*precision*/)
{
  itsLog->origin(LogOrigin("componentlist", "setrefdir"));

  // TODO : IMPLEMENT ME HERE !
  std::string rstat("");
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "getrefdirdec not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::string componentlist::getrefdirframe(const long which)
{
  itsLog->origin(LogOrigin("componentlist", "setrefdir"));

  std::string rstat("");
  try{
    if(itsList && itsBin){
      _checkIndex(which);
      casacore::MDirection md = itsList->component(which).shape().refDirection();
      rstat = md.getRefString().c_str();
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::setrefdir(const long which, const ::casac::variant& ra, 
                              const ::casac::variant& dec, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "setrefdir"));

  bool rstat(false);
  try {
        if(itsList && itsBin){
	 String raval = ra.toString();
	 String decval = dec.toString();
         Quantum<Double> myra=casaQuantity(ra);
         Quantum<Double> mydec=casaQuantity(dec);
 
         const MVDirection newDir(myra, mydec);

         const Vector<Int> intVec(_checkIndices(which, "setrefdir",
                                        "Direction not changed on any components"));
         itsList->setRefDirection(intVec, newDir);
         rstat = true;
	} else {
	  *itsLog << LogIO::WARN << "componentlist is not opened, please open first" << LogIO::POST;
	}
  } catch (const AipsError& x){
	  *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
  }
  return rstat;
}

bool componentlist::setrefdirframe(const long which, const std::string& frame,
                                   const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "setrefdirframe"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
	    casacore::MDirection::Types newFrame;
      if (!casacore::MDirection::getType(newFrame, frame)) {
        *itsLog << LogIO::SEVERE
                << "Could not parse the 'frame' string: Direction frame not changed"
                << LogIO::POST;
      }
      Vector<Int> intVec(1,which);
      itsList->setRefDirectionFrame(intVec, newFrame);
      rstat = true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (AipsError x){
    *itsLog << LogIO::SEVERE
            << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::convertrefdir(const long which, const std::string& frame)
{
  itsLog->origin(LogOrigin("componentlist", "convertrefdir"));

  bool rstat(false);
  try {
    if(itsList && itsBin){
	    casacore::MDirection::Types newFrame;
      if (!casacore::MDirection::getType(newFrame, frame)) {
        *itsLog << LogIO::SEVERE
                << "Could not parse the 'frame' string: Direction frame not changed"
                << LogIO::POST;
        return false;
      }                               
      Vector<Int> intVec(1,which);
      itsList->convertRefDirection(intVec, newFrame);
      rstat = true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } 
  catch (const AipsError& x){
	  *itsLog << LogIO::SEVERE
                  << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
  }
  return rstat;
}

std::string
componentlist::shapetype(const long which)
{
  itsLog->origin(LogOrigin("componentlist", "shapetype"));

  std::string rstat("");
  try {
    if(itsList && itsBin){
      rstat = ComponentType::name(itsList->component(which).shape().type()).c_str();
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x) {
    *itsLog << LogIO::SEVERE
            << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::fromrecord(const ::casac::record& rec)
{
  itsLog->origin(LogOrigin("componentlist", "fromrecord"));

  bool rstat=false;
  try {
    if(itsList && (itsList->nelements() > 0)){
      *itsLog << LogIO::WARN
              << "componentlist is not empty; overwriting it" << LogIO::POST;
      delete itsList;
      itsList= new ComponentList();
      if(!itsBin){
	delete itsBin;
      }
      itsBin= new ComponentList();
    }
    Record *elRec=toRecord(rec);
    String error;
    if(!(itsList->fromRecord(error, *elRec))){
      delete elRec;
      throw(AipsError(String("Error ")+error+ String(" in converting from record")));
    }
    delete elRec;
    rstat=true;

  } catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

::casac::record* componentlist::torecord()
{
  itsLog->origin(LogOrigin("componentlist", "torecord"));

  casac::record* rstat(0);
  try {
    if(itsList && (itsList->nelements() > 0)){
      Record theRec;
      String error;
      if(!(itsList->toRecord(error, theRec))){
        throw(AipsError("converting componentlist to record failed with "+error));
      }
      rstat = fromRecord(theRec); 
    }
    else{
      *itsLog << LogIO::WARN << "componentlist is empty" << LogIO::POST;

    }
       
  } catch (const AipsError& x){
	  *itsLog << LogIO::SEVERE
                  << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
	    }
  return rstat;
}

::casac::record* componentlist::getshape(const long which)
{
  itsLog->origin(LogOrigin("componentlist", "getshape"));

  casac::record* rstat(0);
  try {
    if(itsList && itsBin){
      const ComponentShape& shape = itsList->component(which).shape();
      Record rec;
      String errorMessage;
      if (!shape.toRecord(errorMessage, rec)) {
        *itsLog << "Could not get the component shape because:" << endl
                << errorMessage
                << "Empty record returned" << LogIO::EXCEPTION;
      }
      rec.removeField(RecordFieldId("type"));
      rec.removeField(RecordFieldId("direction"));
      rstat = fromRecord(rec);
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
	  *itsLog << LogIO::SEVERE
                  << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
  }
  return rstat;
}

bool componentlist::setshape(const long which, const std::string& type,
                             const ::casac::variant& majoraxis,
                             const ::casac::variant& minoraxis,
                             const ::casac::variant& positionangle,
                             const ::casac::variant& majoraxiserror,
                             const ::casac::variant& minoraxiserror,
                             const ::casac::variant& positionangleerror,
                             const std::vector<double>& /*optionalparms*/,
                             const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "setshape"));

  bool rstat(false);
  try {
    if(itsList && itsBin){
      ComponentType::Shape reqShape = ComponentType::shape(type);
      ComponentShape* shapePtr = ComponentType::construct(reqShape);
      if (shapePtr == 0) {
        *itsLog << LogIO::SEVERE
                << "Could not translate the shape type to a known value." << endl
                << "Known types are:" << endl;
        for (uInt i = 0; i < ComponentType::NUMBER_SHAPES - 1; i++) {
          reqShape = (ComponentType::Shape) i;
          if (reqShape!=(ComponentType::Shape) 3)
            *itsLog <<  ComponentType::name(reqShape) + String("\n");
        }
        *itsLog << "Shape not changed." << LogIO::POST;
        return false;
      }
      else {
        if (reqShape == ComponentType::Shape::LDISK) {
        *itsLog << LogIO::SEVERE
                << ComponentType::name(reqShape) 
                << " is not implemented yet." << endl;
        *itsLog << "Shape not changed." << LogIO::POST;
        return false;
        }
      }

       
      String errorMessage;
      Record rec;
      Record ma1;
      QuantumHolder(casaQuantity(majoraxis)).toRecord(errorMessage, ma1);
      Record ma2;
      QuantumHolder(casaQuantity(minoraxis)).toRecord(errorMessage, ma2);
      Record ma3;
      QuantumHolder(casaQuantity(positionangle)).toRecord(errorMessage, ma3);
      rec.defineRecord("majoraxis",ma1);
      rec.defineRecord("minoraxis",ma2);
      rec.defineRecord("positionangle",ma3);
       
      if( !(String(majoraxiserror.toString())==String("[]")) &&
          !(String(minoraxiserror.toString())==String("[]")) &&
          !(String(positionangleerror.toString())==String("[]"))){
        Record ma4;
        QuantumHolder(casaQuantity(majoraxiserror)).toRecord(errorMessage, ma4);
        Record ma5;
        QuantumHolder(casaQuantity(minoraxiserror)).toRecord(errorMessage, ma5);
        Record ma6;
        QuantumHolder(casaQuantity(positionangleerror)).toRecord(errorMessage, ma6);
       
        rec.defineRecord("majoraxiserror",ma4);
        rec.defineRecord("minoraxiserror",ma5);
        rec.defineRecord("positionangleerror",ma6);
      }
      else{
        Record ma4;
        QuantumHolder(casacore::Quantity(0.0,"rad")).toRecord(errorMessage, ma4);
        rec.defineRecord("majoraxiserror",ma4);
        rec.defineRecord("minoraxiserror",ma4);
        rec.defineRecord("positionangleerror",ma4);
      }
      if (!shapePtr->fromRecord(errorMessage, rec)) {
        *itsLog << LogIO::SEVERE
                << "Could not parse the shape parameters. The error was:" << endl
                << "\t" << errorMessage
                << "Shape not changed." << LogIO::POST;
        return false;
      }
      Vector<Int> intVec(1,which);
      // Uncomment when limb darkened disk shape is fully supported 
      //if (type=="limbdarkeneddisk") {
      //  shapePtr->setOptParameters(optionalparms);
      //  itsList->setOptParms(intVec, *shapePtr);
      //}
      itsList->setShapeParms(intVec, *shapePtr);

      delete shapePtr;
      rstat = true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
	  *itsLog << LogIO::SEVERE
                  << "Exception Reported: " << x.getMesg() << LogIO::POST;
	  RETHROW(x)
  }
  return rstat;
}

bool componentlist::convertshape(const long /*which*/, const std::string& /*majoraxis*/,
                                 const std::string& /*minoraxis*/,
                                 const std::string& /*positionangle*/)
{
  itsLog->origin(LogOrigin("componentlist", "convertshape"));
  
  // TODO : IMPLEMENT ME HERE !       
  bool rstat(false);
  try {
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "convertshape not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::string componentlist::spectrumtype(const long which)
{
  itsLog->origin(LogOrigin("componentlist", "spectrumtype"));

  std::string rstat("");
  try {
    if(itsList && itsBin){
      rstat = ComponentType::name(itsList->component(which).spectrum().type()).c_str();
    }
    else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x) {
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

::casac::record* componentlist::getspectrum(const long which)
{
  itsLog->origin(LogOrigin("componentlist", "getspectrum"));

  casac::record *rstat(0);
  try{
    if(itsList && itsBin){
      const SpectralModel& spectrum = itsList->component(which).spectrum();
      String errorMessage;
      Record rec;
      if (!spectrum.toRecord(errorMessage, rec)) {
        *itsLog << "Could not get the component spectrum because:" << endl
                << errorMessage
                << "Empty record returned" << LogIO::EXCEPTION;
      }
      rstat = fromRecord (rec);
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::setspectrum(
    long which, const std::string& eltype, const variant& index,
    const std::vector<double>& tabfreqs,
    const std::vector<double>& tabflux, const std::string& freqframe
) {
    itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
    try {
        if(! (itsList && itsBin)) {
            *itsLog << LogIO::WARN
                << "componentlist is not opened, please open first" << LogIO::POST;
            return false;
        }
        String type(eltype);
        type.upcase();
        unique_ptr<SpectralModel> spectrumPtr;
        if (type.startsWith("T")) {
            ThrowIf(
                tabfreqs.size() < 2, 
                "There need to be at least 2 points in a tabular spectrum"
            );
	        ThrowIf(
                tabfreqs.size() != tabflux.size(),
	            "lengths of tabular frequencies and values have to be the same"
            );
	        Vector<MVFrequency> freqs(tabfreqs.size());
	        Vector<Flux<Double>> fluxval(tabfreqs.size());
	        for (uInt i=0; i<uInt(tabfreqs.size()); ++i) {
	            freqs[i] = casacore::Quantity(tabfreqs[i], "Hz");
	            fluxval[i] = Flux<Double>(tabflux[i]);
	        }
	        MFrequency::Types freqFrameType;
	        ThrowIf(
                ! MFrequency::getType(freqFrameType, freqframe),
	            freqframe + " is an unsupported frequency frame"
            );
	        MFrequency refreq(freqs[0], freqFrameType);
	        spectrumPtr.reset(
                new TabularSpectrum(refreq, freqs, fluxval, MFrequency::Ref(freqFrameType))
            );
        }
        else if (type.startsWith("S")) {
            auto indexType = index.type();
            double indexParam = 0;
            if (indexType == variant::DOUBLE || indexType == variant::INT) {
                indexParam = index.toDouble();
            }
            else if (indexType == variant::DOUBLEVEC || indexType == variant::INTVEC) {
                indexParam = index.toDoubleVec()[0];
            }
            else {
                ThrowCc("Unrecognized type for index, must be double or doubleArray");
            }
            MFrequency refFreq = itsList->component(which).spectrum().refFrequency();
            spectrumPtr.reset(new SpectralIndex(refFreq, indexParam));
        }
        else if (type.startsWith("P")) {
            auto indexType = index.type();
            ThrowIf(
                indexType != variant::DOUBLEVEC && indexType != variant::INTVEC,
                "for a spectrum='plp', index must be a list of numbers"
            );
            MFrequency refFreq = itsList->component(which).spectrum().refFrequency();
            Vector<Double> const indexV(index.toDoubleVec());
            spectrumPtr.reset(new PowerLogPoly(refFreq, indexV));
        }
        else if (type.startsWith("C")) {
            spectrumPtr.reset(new ConstantSpectrum());
            MFrequency refFreq = itsList->component(which).spectrum().refFrequency();
            spectrumPtr->setRefFrequency(refFreq);
        }
        else {
            ThrowCc("Unkinown spectral type " + eltype);
        }
        Vector<Int> intVec(1, which);
        itsList->setSpectrumParms(intVec, *spectrumPtr);
        return true;
    }
    catch (const AipsError& x) {
        *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
        RETHROW(x)
    }
    return false;
}

bool componentlist::setstokesspectrum(const long which, const std::string& eltype,
				      const std::vector<double>& index, const std::vector<double>& tabfreqs, const std::vector<double>& tabi, const std::vector<double>& tabq, 
				      const std::vector<double>& tabu,  const std::vector<double>& tabv, const ::casac::variant& reffreq,  
				      const std::string& freqframe)
{
  itsLog->origin(LogOrigin("componentlist", __FUNCTION__));


  bool rstat(false);
  try {
    if(itsList && itsBin){
      String type(eltype);
      type.upcase();
      if(type.contains("TABU"))
	 type="Tabular Spectrum";	 
      if(type.contains("SPECTRAL"))
	 type="Spectral Index";
       ComponentType::SpectralShape reqSpectrum = ComponentType::spectralShape(type);
       SpectralModel* spectrumPtr = ComponentType::construct(reqSpectrum);
       if (spectrumPtr == 0) {
         *itsLog << LogIO::SEVERE
                << "Could not translate the spectral type to a known value." << endl
                << "Known types are:" << endl;
         for(uInt i = 0; i < ComponentType::NUMBER_SPECTRAL_SHAPES - 1; ++i){
           reqSpectrum = static_cast<ComponentType::SpectralShape>(i);
           *itsLog << "\t" << ComponentType::name(reqSpectrum) + String("\n");
         } 
         *itsLog << "Spectrum not changed." << LogIO::POST;
         return false;
       }
       if(reqSpectrum==ComponentType::TABULAR_SPECTRUM){
	 if(tabfreqs.size() <2)
	   throw(AipsError("There need to be at least 2 points in a tabular spectrum to interpolate in between"));
	 if(tabfreqs.size() != tabi.size())
	   throw(AipsError("lengths of tabular frequencies and I values have to be the same"));
	 if(tabfreqs.size() != tabq.size())
	   throw(AipsError("lengths of tabular frequencies and Q values have to be the same"));
	 if(tabfreqs.size() != tabu.size())
	   throw(AipsError("lengths of tabular frequencies and U values have to be the same"));
	  if(tabfreqs.size() != tabv.size())
	   throw(AipsError("lengths of tabular frequencies and V values have to be the same"));
	 Vector<MVFrequency> freqs(tabfreqs.size());
	 Vector<Flux<Double> > fluxval(tabfreqs.size());
	 for (Int i=0; i < Int(tabfreqs.size()) ; ++i){
	   freqs[i]=casacore::Quantity(tabfreqs[i], "Hz");
	   fluxval[i]=Flux<Double>(tabi[i], tabq[i], tabu[i], tabv[i]);
	 }
	 MFrequency::Types freqFrameType;
	 if(!MFrequency::getType(freqFrameType, freqframe))
	   throw(AipsError(String(freqframe) + String(" is not a frequency frame that is understood")));
	 MFrequency refreq(freqs[0], freqFrameType);
	 delete spectrumPtr;
	 spectrumPtr=new TabularSpectrum(refreq, freqs, fluxval, MFrequency::Ref(freqFrameType));
	 

       }
       else{
	 String errorMessage;
	 Record rec;
	 Record freqRec;
	 
	 casacore::Quantity freqQ=String(reffreq.toString()) != String("[]") ? casaQuantity(reffreq) : casacore::Quantity(1.4, "GHz");
	 MFrequency::Types eltype;
	 if(!MFrequency::getType(eltype, freqframe))
	   throw(AipsError(String(freqframe) + String(" is not a frequency frame that is understood")));
	 MeasureHolder(MFrequency(freqQ, eltype)).toRecord(errorMessage,freqRec);
	  setfreq(which, freqQ.get("GHz").getValue(), "GHz", true);
	  setfreqframe(which, freqframe, true);
	 rec.defineRecord("frequency", freqRec);
	 *itsLog << LogIO::DEBUG1 << "index: " << index << LogIO::POST;
	 // Vector<Double> indexVec(index);
	 // if(indexVec.nelements() > 0)
	 //   rec.define("index", indexVec[0]);
	 // else
	 rec.define("index", index[0]);
	 if(index.size() == 4)
	   rec.define("stokesindex", Vector<Double>(index));
	 if (!spectrumPtr->fromRecord(errorMessage, rec)) {
	   *itsLog << LogIO::SEVERE
		   << "Could not parse the spectrum parameters. The error was:" << endl
		   << "\t" << errorMessage << endl
		   << "Spectrum not changed."
		   << LogIO::POST;
	   return false;
	 }
       }
       Vector<Int> intVec(1, which);
       itsList->setSpectrumParms(intVec, *spectrumPtr);
       delete spectrumPtr;

       rstat = true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

::casac::record* componentlist::getfreq(const long /*which*/)
{
  itsLog->origin(LogOrigin("componentlist", "getfreq"));

  // TODO : IMPLEMENT ME HERE !
  ::casac::record *rstat(0);
  try {
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "getfreq not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

double componentlist::getfreqvalue(const long /*which*/)
{
  itsLog->origin(LogOrigin("componentlist", "getfreqvalue"));

  // TODO : IMPLEMENT ME HERE !
  double rstat(false);
  try {
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "getfreqvalue not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::string componentlist::getfrequnit(const long /*which*/)
{
  itsLog->origin(LogOrigin("componentlist", "getfrequnit"));

  // TODO : IMPLEMENT ME HERE !
  std::string rstat("");
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "getfrequnit not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

std::string componentlist::getfreqframe(const long /*which*/)
{
  itsLog->origin(LogOrigin("componentlist", "getfreqframe"));

  // TODO : IMPLEMENT ME HERE !
  std::string rstat("");
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "getfreqframe not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch(const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::setfreq(const long which, const double value,
                            const std::string& unit, const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "setfreq"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      const MVFrequency newFreq(Quantum<Double>(value, Unit(unit)));
      Vector<Int> intVec(1,which);
      itsList->setRefFrequency(intVec, newFreq);
      convertfrequnit(which, unit);
		    
      rstat = true;
    }
    else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::setfreqframe(const long which, const std::string& frame,
                                 const bool /*log*/)
{
  itsLog->origin(LogOrigin("componentlist", "setfreqframe"));
  
  bool rstat(false);
  try{
    if(itsList && itsBin){
      MFrequency::Types newFrame;
      if (!MFrequency::getType(newFrame, frame)) {
        *itsLog << LogIO::SEVERE
                << "Could not parse the 'frame' string: Frequency frame not changed"
                << LogIO::POST;
        return false;
      }
      Vector<Int> intVec(1, which);
      itsList->setRefFrequencyFrame(intVec, newFrame);

      rstat = true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  } catch (const AipsError& x){
    *itsLog << LogIO::SEVERE
            << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::convertfrequnit(const long which, const std::string& unit)
{
  itsLog->origin(LogOrigin("componentlist", "convertfrequnit"));
  
  bool rstat(false);
  try{
    if(itsList && itsBin){
      Vector<Int> intVec(1, which);
      itsList->setRefFrequencyUnit(intVec, Unit(unit));
      rstat = true;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}


::casac::record* componentlist::getcomponent(const long which, const bool /*iknow*/)
{
  itsLog->origin(LogOrigin("componentlist", "getcomponent"));

  casac::record *rstat(0);
  try{
    if(itsList && itsBin){
      _checkIndex(which);
      const SkyComponent& listRef(itsList->component(which));
      String error;
      Record leRec;
      if(!listRef.toRecord(error, leRec)){
        throw(AipsError(String("Converting compenent torecord failed with error")
                        + error));
      }
      rstat=fromRecord(leRec);
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE
            << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::add(const ::casac::record& thecomponent, const bool /*iknow*/)
{
  itsLog->origin(LogOrigin("componentlist", "add"));

  bool rstat(false);
  try{
    if(itsList && itsBin){
      Record *elRec=toRecord(thecomponent);
      SkyComponent elComp;
      String err;
      if(!elComp.fromRecord(err, *elRec)){
        throw(AipsError(String("Converting Component from record failed with error ")
                        + err));
      }
      itsList->add(elComp);
      if(elRec !=0){
        delete elRec;
      }
      rstat=true;
    }
    else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::replace(const long /*which*/, const ::casac::record& /*list*/,
                            const std::vector<long>& /*whichones*/)
{
  itsLog->origin(LogOrigin("componentlist", "replace"));

  // TODO : IMPLEMENT ME HERE !
  bool rstat(false);
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN << "replace not implemented yet" << LogIO::POST;
    } else {
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch (const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

bool componentlist::summarize(const long which) {
	itsLog->origin(LogOrigin("componentlist", __FUNCTION__));
	Bool rstat = false;
	try{
		if(itsList && itsBin) {
			if(uInt(which) >= itsList->nelements() ) {
				ostringstream oss;
				oss << "List has only " << itsList->nelements()
    					<< " components, but zero-based component " << which << " specified. "
    					<< "Please specify a component less than " << itsList->nelements();
					throw AipsError(oss.str());
			}
			*itsLog << LogIO::NORMAL << itsList->summarize(which) << LogIO::POST;
			rstat = true;
		} else {
			*itsLog << LogIO::WARN
					<< "componentlist is not opened, please open first" << LogIO::POST;
			rstat = false;
		}
	}
	catch (const AipsError& x){
		*itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
		RETHROW(x)
	}
	return rstat;
}

bool componentlist::iscomponentlist(const ::casac::variant& /*tool*/)
{
  itsLog->origin(LogOrigin("componentlist", "iscomponentlist"));

  // TODO : IMPLEMENT ME HERE !
  bool rstat(false);
  try{
    if(itsList && itsBin){
      *itsLog << LogIO::WARN
              << "is_componentlist not implemented yet" << LogIO::POST;
    }
    else{
      *itsLog << LogIO::WARN
              << "componentlist is not opened, please open first" << LogIO::POST;
    }
  }
  catch(const AipsError& x){
    *itsLog << LogIO::SEVERE << "Exception Reported: " << x.getMesg() << LogIO::POST;
    RETHROW(x)
  }
  return rstat;
}

vector<int> componentlist::_checkIndices(int which, const String& function,
                                         const String& message) const
{
  std::vector<int> whichVec(1, which);
  return _checkIndices(whichVec, function, message);
}
//
//
vector<int> componentlist::_checkIndices(const vector<int>& which,
                                        const String& function,
                                        const String& message) const
{
  itsLog->origin(LogOrigin("componentlist", function));
  
  Vector<int> intVec(which);
  if((intVec.nelements()==1) && (intVec[0]==-1)){
    intVec.resize( itsList->nelements());
    indgen(intVec);
  }
  const Int listLength = itsList->nelements();
  const uInt whichLength = which.size();
  for (uInt c = 0; c < whichLength; c++) {
    if (intVec(c) < 0 || intVec(c) >= listLength) {
      *itsLog << LogIO::SEVERE
              << "Index out of range." << endl
              << "A component number is less than one or greater than"
              << " the list length" << endl
              << message
              << LogIO::EXCEPTION;
    }
  }
  return vector<int>(intVec.begin(),intVec.end());
}

int componentlist::_checkFluxPol(const String& polString)
{
  itsLog->origin(LogOrigin("componentlist", __FUNCTION__));

  const ComponentType::Polarisation pol(ComponentType::polarisation(polString));
  if(pol == ComponentType::UNKNOWN_POLARISATION){
    *itsLog << "Unknown polarization. Possible values are:" << endl;
    for (uInt i = 0; i < ComponentType::NUMBER_POLARISATIONS - 1; i++) {
      *itsLog << " '"
             <<ComponentType::name(static_cast<ComponentType::Polarisation>(i))
             << "' ";
    }
    *itsLog << LogIO::EXCEPTION;
  }
  return pol;
}

void componentlist::_checkIndex(int which) const {
	Int m = itsList->nelements() - 1;
	ThrowIf(
		which < 0 || which > m,
		"Index out of range. The component number must be between 0 and " + String::toString(m)
    );
}

} // casac namespace

